'use strict';
require('./bootstrap');

describe('ORM', function() {

  let Post = ORM.models.post;
  let User = ORM.models.user;
  let Tag = ORM.models.tag;
  let Media = ORM.models.media;

  //
  // Create
  //

  it('should create', function(done) {
    User.create({
      email: 'test@gmail.com'
    })
    .then((model) => {
      model.email.should.equal('test@gmail.com');
      model.should.have.property('created');
      model.created.should.equal(true);
      done();
    });
  });

  //
  // FindOrCreate
  //

  it('should create with findOrCreate', function(done) {
    User.findOrCreate({
      email: 'findorcreateme@gmail.com'
    })
    .then((model) => {
      model.email.should.equal('findorcreateme@gmail.com');
      model.should.have.property('created');
      model.created.should.equal(true);
      done();
    });
  });

  it('should find with findOrCreate', function(done) {
    User.findOrCreate({
      email: 'findorcreateme@gmail.com'
    })
    .then((model) => {
      model.email.should.equal('findorcreateme@gmail.com');
      model.should.not.have.property('created');
      done();
    });
  });

  //
  // Include
  //

  let userA, userB, tagA, tagB, postA, postB, media;

  it('should deep include', function(done) {

    Promise.resolve()

    //
    // Create User
    .then(() => {
      return User.create({
        email: 'usera@gmail.com'
      });
    })
    .then((model) => {
      userA = model;
      return;
    })
    .then(() => {
      return User.create({
        email: 'userb@gmail.com'
      });
    })
    .then((model) => {
      userB = model;
      return;
    })

    //
    // Create Tags
    .then((model) => {
      return Tag.create({
        name: 'taga'
      });
    })
    .then((model) => {
      tagA = model;
      return;
    })
    .then((model) => {
      return Tag.create({
        name: 'tagb'
      });
    })
    .then((model) => {
      tagB = model;
      return;
    })

    //
    // Create Media
    .then(() => {
      return Media.create({
        type: 'image'
      });
    })
    .then((model) => {
      media = model;
      return;
    })

    //
    // Create Post
    .then(() => {
      return Post.create({
        title: 'PostA',
        user: userA,
        tags: [tagA],
        media: media
      });
    })
    .then((model) => {
      postA = model;
      return;
    })
    .then(() => {
      return Post.create({
        title: 'PostB',
        user: userB,
        tags: [tagB],
        media: media
      });
    })
    .then((model) => {
      postB = model;
      return;
    })

    //
    // Find
    .then(() => {
      return User.findOne({ email: 'usera@gmail.com' }, {
        include: [
          {
            relation: 'posts',
            include: [
              { relation: 'tags' },
              { relation: 'media'}
            ]
          }
        ]
      });
    })
    .then((model) => {
      model.id.should.equal(userA.id);
      model.posts.should.have.length(1);
      model.posts[0].id.should.equal(postA.id);
      model.posts[0].tags.should.have.length(1);
      model.posts[0].tags[0].id.should.equal(tagA.id);
      model.posts[0].media.id.should.equal(media.id);
      done();
    });

  });

  //
  // Find
  //

  it('should deep find', function(done) {

    User.find({
      posts: {
        title: 'PostA',
        /*
        // TODO: nth level find
        tags: [
          { name: 'travel'}
        ]
        */
      }
    })
    .then((model) => {
      done();
    })
    .catch((error) => {
      console.log(error);
    })

  });

  //
  // Find Operators
  //

  it('should find all not equal to', function(done) {

    Post.find({
      tags: {
        name: { $ne: 'taga' }
      }
    })
    .then((models) => {
      models.should.have.length(1);
      done();
    })
    .catch((error) => {
      console.log(error);
    })

  });

  it('should find all like', function(done) {

    Post.find({
      tags: {
        name: { $like: 'tag%' }
      }
    })
    .then((models) => {
      models.should.have.length(2);
      done();
    })
    .catch((error) => {
      console.log(error);
    })

  });

  it('should find all or', function(done) {

    Post.find({
      tags: {
        name: { $or: ['taga', 'tagb'] }
      }
    })
    .then((models) => {
      models.should.have.length(2);
      done();
    })
    .catch((error) => {
      console.log(error);
    })

  });

  it('should find all and', function(done) {

    Post.find({
      tags: {
        name: { $and: ['taga'] }
      }
    })
    .then((models) => {
      models.should.have.length(1);
      done();
    })
    .catch((error) => {
      console.log(error);
    })

  });

  //
  // Find Order
  //

  it('should find all by order', function(done) {

    let conditions = {
      title: { $like: 'Post%' },
      tags: {
        name: { $or: ['taga', 'tagb']}
      }
    };

    let include = [
      { relation: 'tags', include: [{ relation: 'posts' }] }
    ];

    Post.find(conditions, {
      include: include,
      order: ['created_at', 'asc']
    })
    .then((models) => {
      models[0].id.should.equal(postA.id);
      models[1].id.should.equal(postB.id);
      return Post.find(conditions, {
        include: include,
        order: ['created_at', 'desc']
      })
    })
    .then((models) => {
      models[0].id.should.equal(postB.id);
      models[1].id.should.equal(postA.id);
      done();
    })
    .catch((error) => {
      console.log(error);
    })

  });

  //
  // Update
  //

  it('should update', function(done) {
    User.update({
      email: 'usera@gmail.com'
    }, {
      email: 'updated@gmail.com'
    })
    .then((models) => {
      models[0].email.should.equal('updated@gmail.com');
      return User.findOne({ email: 'updated@gmail.com' })
    })
    .then((model) => {
      model.email.should.equal('updated@gmail.com');
      done();
    });
  });

  //
  // Delete
  //

  it('should delete', function(done) {
    User.delete({
      email: 'updated@gmail.com'
    })
    .then((models) => {
      models[0].email.should.equal('updated@gmail.com');
      done();
    })
  });

});